@@ -1,7 +1,4 @@
-from . import base
-from . import login
-from . import metrics
-from . import pages
+from . import base, login, metrics, pages
 from .base import *
 from .login import *
 
