@@ -23,32 +23,33 @@ import signal
 import time
 from functools import wraps
 from subprocess import Popen
-from urllib.parse import quote
+from urllib.parse import quote, urlparse
 from weakref import WeakKeyDictionary
 
-from tornado.httpclient import AsyncHTTPClient
-from tornado.httpclient import HTTPError
-from tornado.httpclient import HTTPRequest
+from tornado.httpclient import AsyncHTTPClient, HTTPError, HTTPRequest
 from tornado.ioloop import PeriodicCallback
-from traitlets import Any
-from traitlets import Bool
-from traitlets import default
-from traitlets import Dict
-from traitlets import Instance
-from traitlets import Integer
-from traitlets import observe
-from traitlets import Unicode
+from traitlets import (
+    Any,
+    Bool,
+    CaselessStrEnum,
+    Dict,
+    Instance,
+    Integer,
+    TraitError,
+    Unicode,
+    default,
+    observe,
+    validate,
+)
 from traitlets.config import LoggingConfigurable
 
-from . import utils
-from .metrics import CHECK_ROUTES_DURATION_SECONDS
-from .metrics import PROXY_POLL_DURATION_SECONDS
-from .objects import Server
-from .utils import AnyTimeoutError
-from .utils import exponential_backoff
-from .utils import url_path_join
 from jupyterhub.traitlets import Command
 
+from . import utils
+from .metrics import CHECK_ROUTES_DURATION_SECONDS, PROXY_POLL_DURATION_SECONDS
+from .objects import Server
+from .utils import AnyTimeoutError, exponential_backoff, url_escape_path, url_path_join
+
 
 def _one_at_a_time(method):
     """decorator to limit an async method to be called only once
@@ -122,7 +123,8 @@ class Proxy(LoggingConfigurable):
     )
 
     extra_routes = Dict(
-        {},
+        key_trait=Unicode(),
+        value_trait=Unicode(),
         config=True,
         help="""
         Additional routes to be maintained in the proxy.
@@ -141,6 +143,51 @@ class Proxy(LoggingConfigurable):
         """,
     )
 
+    @validate("extra_routes")
+    def _validate_extra_routes(self, proposal):
+        extra_routes = {}
+        # check routespecs for leading/trailing slashes
+        for routespec, target in proposal.value.items():
+            if not isinstance(routespec, str):
+                raise TraitError(
+                    f"Proxy.extra_routes keys must be str, got {routespec!r}"
+                )
+            if not isinstance(target, str):
+                raise TraitError(
+                    f"Proxy.extra_routes values must be str, got {target!r}"
+                )
+            if not routespec.endswith("/"):
+                # trailing / is unambiguous, so we can add it
+                self.log.warning(
+                    f"Adding missing trailing '/' to c.Proxy.extra_routes {routespec} -> {routespec}/"
+                )
+                routespec += "/"
+
+            if self.app.subdomain_host:
+                # subdomain routing must _not_ start with /
+                if routespec.startswith("/"):
+                    raise ValueError(
+                        f"Proxy.extra_routes missing host component in {routespec} (must not have leading '/') when using `JupyterHub.subdomain_host = {self.app.subdomain_host!r}`"
+                    )
+
+            else:
+                # no subdomains, must start with /
+                # this is ambiguous with host routing, so raise instead of warn
+                if not routespec.startswith("/"):
+                    raise ValueError(
+                        f"Proxy.extra_routes routespec {routespec} missing leading '/'."
+                    )
+
+            # validate target URL?
+            target_url = urlparse(target.lower())
+            if target_url.scheme not in {"http", "https"} or not target_url.netloc:
+                raise ValueError(
+                    f"Proxy.extra_routes target {routespec}={target!r} doesn't look like a URL (should have http[s]://...)"
+                )
+            extra_routes[routespec] = target
+
+        return extra_routes
+
     def start(self):
         """Start the proxy.
 
@@ -306,7 +353,9 @@ class Proxy(LoggingConfigurable):
         """Remove a user's server from the proxy table."""
         routespec = user.proxy_spec
         if server_name:
-            routespec = url_path_join(user.proxy_spec, server_name, '/')
+            routespec = url_path_join(
+                user.proxy_spec, url_escape_path(server_name), '/'
+            )
         self.log.info("Removing user %s from proxy (%s)", user.name, routespec)
         await self.delete_route(routespec)
 
@@ -475,7 +524,21 @@ class ConfigurableHTTPProxy(Proxy):
     def _concurrency_changed(self, change):
         self.semaphore = asyncio.BoundedSemaphore(change.new)
 
+    # https://github.com/jupyterhub/configurable-http-proxy/blob/4.5.1/bin/configurable-http-proxy#L92
+    log_level = CaselessStrEnum(
+        ["debug", "info", "warn", "error"],
+        "info",
+        help="Proxy log level",
+        config=True,
+    )
+
     debug = Bool(False, help="Add debug-level logging to the Proxy.", config=True)
+
+    @observe('debug')
+    def _debug_changed(self, change):
+        if change.new:
+            self.log_level = "debug"
+
     auth_token = Unicode(
         help="""The Proxy auth token
 
@@ -671,11 +734,11 @@ class ConfigurableHTTPProxy(Proxy):
             str(api_server.port),
             '--error-target',
             url_path_join(self.hub.url, 'error'),
+            '--log-level',
+            self.log_level,
         ]
         if self.app.subdomain_host:
             cmd.append('--host-routing')
-        if self.debug:
-            cmd.extend(['--log-level', 'debug'])
         if self.ssl_key:
             cmd.extend(['--ssl-key', self.ssl_key])
         if self.ssl_cert:
