@@ -8,14 +8,11 @@ import pytest
 from pytest import mark
 from tornado.log import app_log
 
-from .. import orm
-from .. import roles
-from ..scopes import get_scopes_for
-from ..scopes import scope_definitions
+from .. import orm, roles
+from ..scopes import get_scopes_for, scope_definitions
 from ..utils import utcnow
 from .mocking import MockHub
-from .utils import add_user
-from .utils import api_request
+from .utils import add_user, api_request
 
 
 @mark.role
@@ -76,24 +73,17 @@ def test_orm_roles(db):
     # assigns it the default 'token' role
     token = user.new_api_token()
     user_token = orm.APIToken.find(db, token=token)
-    assert user_token in token_role.tokens
-    assert token_role in user_token.roles
+    assert set(user_token.scopes) == set(token_role.scopes)
 
     # check creating token with a specific role
     token = service.new_api_token(roles=['service'])
     service_token = orm.APIToken.find(db, token=token)
-    assert service_token in service_role.tokens
-    assert service_role in service_token.roles
+    assert set(service_token.scopes) == set(service_role.scopes)
 
-    # check deleting user removes the user and the token from roles
+    # check deleting user removes the user from roles
     db.delete(user)
     db.commit()
     assert user_role.users == []
-    assert user_token not in token_role.tokens
-    # check deleting the service token removes it from 'service' role
-    db.delete(service_token)
-    db.commit()
-    assert service_token not in service_role.tokens
     # check deleting the service_role removes it from service.roles
     db.delete(service_role)
     db.commit()
@@ -174,7 +164,7 @@ def test_orm_roles_delete_cascade(db):
 
 @mark.role
 @mark.parametrize(
-    "scopes, subscopes",
+    "scopes, expected_scopes",
     [
         (
             ['admin:users'],
@@ -246,12 +236,12 @@ def test_orm_roles_delete_cascade(db):
         ),
     ],
 )
-def test_get_subscopes(db, scopes, subscopes):
-    """Test role scopes expansion into their subscopes"""
+def test_get_expanded_scopes(db, scopes, expected_scopes):
+    """Test role scopes expansion into their fully expanded scopes"""
     roles.create_role(db, {'name': 'testing_scopes', 'scopes': scopes})
     role = orm.Role.find(db, name='testing_scopes')
-    response = roles._get_subscopes(role)
-    assert response == subscopes
+    expanded_scopes = roles.roles_to_expanded_scopes([role], owner=None)
+    assert expanded_scopes == expected_scopes
     db.delete(role)
 
 
@@ -443,7 +433,14 @@ async def test_scope_existence(tmpdir, request, role, response):
 
 
 @mark.role
-async def test_load_roles_users(tmpdir, request):
+@mark.parametrize(
+    "explicit_allowed_users",
+    [
+        (True,),
+        (False,),
+    ],
+)
+async def test_load_roles_users(tmpdir, request, explicit_allowed_users):
     """Test loading predefined roles for users in app.py"""
     roles_to_load = [
         {
@@ -461,7 +458,8 @@ async def test_load_roles_users(tmpdir, request):
     hub.init_db()
     db = hub.db
     hub.authenticator.admin_users = ['admin']
-    hub.authenticator.allowed_users = ['cyclops', 'gandalf', 'bilbo', 'gargamel']
+    if explicit_allowed_users:
+        hub.authenticator.allowed_users = ['cyclops', 'gandalf', 'bilbo', 'gargamel']
     await hub.init_role_creation()
     await hub.init_users()
     await hub.init_role_assignment()
@@ -490,7 +488,7 @@ async def test_load_roles_users(tmpdir, request):
 
 
 @mark.role
-async def test_load_roles_services(tmpdir, request):
+async def test_load_roles_services(tmpdir, request, preserve_scopes):
     services = [
         {'name': 'idle-culler', 'api_token': 'some-token'},
         {'name': 'user_service', 'api_token': 'some-other-token'},
@@ -501,6 +499,11 @@ async def test_load_roles_services(tmpdir, request):
         'some-other-token': 'user_service',
         'secret-token': 'admin_service',
     }
+    custom_scopes = {
+        "custom:empty-scope": {
+            "description": "empty custom scope",
+        }
+    }
     roles_to_load = [
         {
             'name': 'idle-culler',
@@ -510,11 +513,13 @@ async def test_load_roles_services(tmpdir, request):
                 'read:users:activity',
                 'read:servers',
                 'servers',
+                'custom:empty-scope',
             ],
             'services': ['idle-culler'],
         },
     ]
     kwargs = {
+        'custom_scopes': custom_scopes,
         'load_roles': roles_to_load,
         'services': services,
         'service_tokens': service_tokens,
@@ -644,9 +649,9 @@ async def test_load_roles_user_tokens(tmpdir, request):
     # test if all other tokens have default 'user' role
     token_role = orm.Role.find(db, 'token')
     secret_token = orm.APIToken.find(db, 'secret-token')
-    assert token_role in secret_token.roles
+    assert set(secret_token.scopes) == set(token_role.scopes)
     secrety_token = orm.APIToken.find(db, 'secrety-token')
-    assert token_role in secrety_token.roles
+    assert set(secrety_token.scopes) == set(token_role.scopes)
 
     # delete the test tokens
     for token in db.query(orm.APIToken):
@@ -665,13 +670,17 @@ async def test_load_roles_user_tokens(tmpdir, request):
         # no role requested - gets default 'token' role
         ({}, None, None, 201),
         # role scopes within the user's default 'user' role
-        ({}, 'self-reader', ['read:users'], 201),
+        ({}, 'self-reader', ['read:users!user'], 201),
+        # role scopes within the user's default 'user' role, but with disjoint filter
+        ({}, 'other-reader', ['read:users!user=other'], 400),
+        # role scopes within the user's default 'user' role, without filter
+        ({}, 'other-reader', ['read:users'], 400),
         # role scopes outside of the user's role but within the group's role scopes of which the user is a member
         ({}, 'groups-reader', ['read:groups'], 201),
         # non-existing role request
-        ({}, 'non-existing', [], 404),
+        ({}, 'non-existing', [], 400),
         # role scopes outside of both user's role and group's role scopes
-        ({}, 'users-creator', ['admin:users'], 403),
+        ({}, 'users-creator', ['admin:users'], 400),
     ],
 )
 async def test_get_new_token_via_api(app, headers, rolename, scopes, status):
@@ -737,7 +746,7 @@ async def test_self_expansion(app, kind, has_user_scopes):
     test_role = orm.Role(name='test_role', scopes=['self'])
     orm_obj.roles.append(test_role)
     # test expansion of user/service scopes
-    scopes = roles.expand_roles_to_scopes(orm_obj)
+    scopes = get_scopes_for(orm_obj)
     assert bool(scopes) == has_user_scopes
     if kind == 'users':
         for scope in scopes:
@@ -778,9 +787,9 @@ async def test_user_filter_expansion(app, scope_list, kind, test_for_token):
     if test_for_token:
         token = orm_obj.new_api_token(roles=['test_role'])
         orm_token = orm.APIToken.find(app.db, token)
-        expanded_scopes = roles.expand_roles_to_scopes(orm_token)
+        expanded_scopes = get_scopes_for(orm_token)
     else:
-        expanded_scopes = roles.expand_roles_to_scopes(orm_obj)
+        expanded_scopes = get_scopes_for(orm_obj)
 
     for scope in scope_list:
         base, _, filter = scope.partition('!')
@@ -797,19 +806,6 @@ async def test_user_filter_expansion(app, scope_list, kind, test_for_token):
     app.db.delete(test_role)
 
 
-async def test_large_filter_expansion(app, create_temp_role, create_user_with_scopes):
-    scope_list = roles.expand_self_scope('==')
-    # Mimic the role 'self' based on '!user' filter for tokens
-    scope_list = [scope.rstrip("=") for scope in scope_list]
-    filtered_role = create_temp_role(scope_list)
-    user = create_user_with_scopes('self')
-    user.new_api_token(roles=[filtered_role.name])
-    user.new_api_token(roles=['token'])
-    manual_scope_set = get_scopes_for(user.api_tokens[0])
-    auto_scope_set = get_scopes_for(user.api_tokens[1])
-    assert manual_scope_set == auto_scope_set
-
-
 @mark.role
 @mark.parametrize(
     "name, valid",
@@ -845,9 +841,7 @@ async def test_server_token_role(app):
     assert orm_server_token
 
     server_role = orm.Role.find(app.db, 'server')
-    token_role = orm.Role.find(app.db, 'token')
-    assert server_role in orm_server_token.roles
-    assert token_role not in orm_server_token.roles
+    assert set(server_role.scopes) == set(orm_server_token.scopes)
 
     assert orm_server_token.user.name == user.name
     assert user.api_tokens == [orm_server_token]
@@ -872,7 +866,7 @@ async def test_server_role_api_calls(
     user = add_user(app.db, app, name='test_user')
     roles.grant_role(app.db, user, 'user')
     app_log.debug(user.roles)
-    app_log.debug(roles.expand_roles_to_scopes(user.orm_user))
+    app_log.debug(get_scopes_for(user.orm_user))
     if token_role == 'no_role':
         api_token = user.new_api_token(roles=[])
     else:
@@ -900,20 +894,18 @@ async def test_server_role_api_calls(
     assert r.status_code == response
 
 
-async def test_oauth_allowed_roles(app, create_temp_role):
-    allowed_roles = ['oracle', 'goose']
+async def test_oauth_client_allowed_scopes(app):
+    allowed_scopes = ['read:users', 'read:groups']
     service = {
         'name': 'oas1',
         'api_token': 'some-token',
-        'oauth_roles': ['oracle', 'goose'],
+        'oauth_client_allowed_scopes': allowed_scopes,
     }
-    for role in allowed_roles:
-        create_temp_role('read:users', role_name=role)
     app.services.append(service)
     app.init_services()
     app_service = app.services[0]
     assert app_service['name'] == 'oas1'
-    assert set(app_service['oauth_roles']) == set(allowed_roles)
+    assert set(app_service['oauth_client_allowed_scopes']) == set(allowed_scopes)
 
 
 async def test_user_group_roles(app, create_temp_role):
@@ -944,7 +936,7 @@ async def test_user_group_roles(app, create_temp_role):
     # regression test for #3472
     roles_before = list(user.roles)
     for i in range(3):
-        roles.expand_roles_to_scopes(user.orm_user)
+        get_scopes_for(user.orm_user)
         user_roles = list(user.roles)
         assert user_roles == roles_before
 
@@ -1046,33 +1038,6 @@ async def test_config_role_users():
     assert role not in user.roles
 
 
-async def test_scope_expansion_revokes_tokens(app, mockservice_url):
-    role_name = 'morpheus'
-    roles_to_load = [
-        {
-            'name': role_name,
-            'description': 'wears sunglasses',
-            'scopes': ['users', 'groups'],
-        },
-    ]
-    app.load_roles = roles_to_load
-    await app.init_role_creation()
-    user = add_user(app.db, name='laurence')
-    for _ in range(2):
-        user.new_api_token()
-    red_token, blue_token = user.api_tokens
-    roles.grant_role(app.db, red_token, role_name)
-    service = mockservice_url
-    red_token.client_id = service.oauth_client_id
-    app.db.commit()
-    # Restart hub and see if token is revoked
-    app.load_roles[0]['scopes'].append('proxy')
-    await app.init_role_creation()
-    user = orm.User.find(app.db, name='laurence')
-    assert red_token not in user.api_tokens
-    assert blue_token in user.api_tokens
-
-
 async def test_duplicate_role_users():
     role_name = 'painter'
     user_name = 'benny'
@@ -1183,14 +1148,47 @@ async def test_no_admin_role_change():
         await hub.init_role_creation()
 
 
-async def test_user_config_respects_memberships():
+@pytest.mark.parametrize(
+    "in_db, role_users, allowed_users, expected_members",
+    [
+        # users in the db, not specified in custom user role
+        # no change to membership
+        (["alpha", "beta"], None, None, ["alpha", "beta"]),
+        # allowed_users is additive, not strict
+        (["alpha", "beta"], None, {"gamma"}, ["alpha", "beta", "gamma"]),
+        # explicit empty revokes all assignments
+        (["alpha", "beta"], [], None, []),
+        # explicit value is respected exactly
+        (["alpha", "beta"], ["alpha", "gamma"], None, ["alpha", "gamma"]),
+    ],
+)
+async def test_user_role_from_config(
+    in_db, role_users, allowed_users, expected_members
+):
+    role_spec = {
+        'name': 'user',
+        'scopes': ['self', 'shutdown'],
+    }
+    if role_users is not None:
+        role_spec['users'] = role_users
+    hub = MockHub(load_roles=[role_spec])
+    hub.init_db()
+    db = hub.db
+    hub.authenticator.admin_users = set()
+    if allowed_users:
+        hub.authenticator.allowed_users = allowed_users
+    await hub.init_role_creation()
+
+
+async def test_user_config_creates_default_role():
     role_spec = [
         {
-            'name': 'user',
-            'scopes': ['self', 'shutdown'],
+            'name': 'new-role',
+            'scopes': ['read:users'],
+            'users': ['not-yet-created-user'],
         }
     ]
-    user_names = ['eddy', 'carol']
+    user_names = []
     hub = MockHub(load_roles=role_spec)
     hub.init_db()
     hub.authenticator.allowed_users = user_names
@@ -1198,9 +1196,9 @@ async def test_user_config_respects_memberships():
     await hub.init_users()
     await hub.init_role_assignment()
     user_role = orm.Role.find(hub.db, 'user')
-    for user_name in user_names:
-        user = orm.User.find(hub.db, user_name)
-        assert user in user_role.users
+    new_role = orm.Role.find(hub.db, 'new-role')
+    assert orm.User.find(hub.db, 'not-yet-created-user') in new_role.users
+    assert orm.User.find(hub.db, 'not-yet-created-user') in user_role.users
 
 
 async def test_admin_role_respects_config():
@@ -1222,16 +1220,45 @@ async def test_admin_role_respects_config():
         assert user in admin_role.users
 
 
-async def test_empty_admin_spec():
-    role_spec = [{'name': 'admin', 'users': []}]
-    hub = MockHub(load_roles=role_spec)
+@pytest.mark.parametrize(
+    "in_db, role_users, admin_users, expected_members",
+    [
+        # users in the db, not specified in custom user role
+        # no change to membership
+        (["alpha", "beta"], None, None, ["alpha", "beta"]),
+        # admin_users is additive, not strict
+        (["alpha", "beta"], None, {"gamma"}, ["alpha", "beta", "gamma"]),
+        # explicit empty revokes all assignments
+        (["alpha", "beta"], [], None, []),
+        # explicit value is respected exactly
+        (["alpha", "beta"], ["alpha", "gamma"], None, ["alpha", "gamma"]),
+    ],
+)
+async def test_admin_role_membership(in_db, role_users, admin_users, expected_members):
+
+    load_roles = []
+    if role_users is not None:
+        load_roles.append({"name": "admin", "users": role_users})
+    if not admin_users:
+        admin_users = set()
+    hub = MockHub(load_roles=load_roles, db_url="sqlite:///:memory:")
     hub.init_db()
-    hub.authenticator.admin_users = []
     await hub.init_role_creation()
+    db = hub.db
+    hub.authenticator.admin_users = admin_users
+    # add in_db users to the database
+    # this is the 'before' state of who had the role before startup
+    for username in in_db or []:
+        user = orm.User(name=username)
+        db.add(user)
+        db.commit()
+        roles.grant_role(db, user, "admin")
+    db.commit()
     await hub.init_users()
     await hub.init_role_assignment()
-    admin_role = orm.Role.find(hub.db, 'admin')
-    assert not admin_role.users
+    admin_role = orm.Role.find(db, 'admin')
+    role_members = sorted(user.name for user in admin_role.users)
+    assert role_members == expected_members
 
 
 async def test_no_default_service_role():
@@ -1271,7 +1298,7 @@ async def test_hub_upgrade_detection(tmpdir):
     for name in user_names:
         user = orm.User.find(hub.db, name)
         assert user_role in user.roles
-        assert token_role in user.api_tokens[0].roles
+        assert set(user.api_tokens[0].scopes) == set(token_role.scopes)
     # Strip all roles and see if it sticks
     user_role.users = []
     token_role.tokens = []
@@ -1290,50 +1317,10 @@ async def test_hub_upgrade_detection(tmpdir):
     allowed_user = orm.User.find(hub.db, 'patricia')
     rem_user = orm.User.find(hub.db, 'quentin')
     assert user_role in allowed_user.roles
-    assert token_role not in allowed_user.api_tokens[0].roles
     assert user_role not in rem_user.roles
     assert token_role not in rem_user.roles
 
 
-async def test_token_keep_roles_on_restart():
-    role_spec = [
-        {
-            'name': 'bloop',
-            'scopes': ['read:users'],
-        }
-    ]
-    hub = MockHub(load_roles=role_spec)
-    hub.init_db()
-    hub.authenticator.admin_users = ['ben']
-    await hub.init_role_creation()
-    await hub.init_users()
-    await hub.init_role_assignment()
-    user = orm.User.find(hub.db, name='ben')
-    for _ in range(3):
-        user.new_api_token()
-    happy_token, content_token, sad_token = user.api_tokens
-    roles.grant_role(hub.db, happy_token, 'bloop')
-    roles.strip_role(hub.db, sad_token, 'token')
-    assert len(happy_token.roles) == 2
-    assert len(content_token.roles) == 1
-    assert len(sad_token.roles) == 0
-    # Restart hub and see if roles are as expected
-    hub.load_roles = []
-    await hub.init_role_creation()
-    await hub.init_users()
-    await hub.init_api_tokens()
-    await hub.init_role_assignment()
-    user = orm.User.find(hub.db, name='ben')
-    happy_token, content_token, sad_token = user.api_tokens
-    assert len(happy_token.roles) == 1
-    assert len(content_token.roles) == 1
-    print(sad_token.roles)
-    assert len(sad_token.roles) == 0
-    for token in user.api_tokens:
-        hub.db.delete(token)
-    hub.db.commit()
-
-
 async def test_login_default_role(app, username):
     cookies = await app.login_user(username)
     user = app.users[username]
@@ -1344,7 +1331,8 @@ async def test_login_default_role(app, username):
     user.roles = []
     app.db.commit()
 
-    # login *again*; user exists, shouldn't trigger change in roles
+    # login *again*; user exists,
+    # login should always trigger "user" role assignment
     cookies = await app.login_user(username)
     user = app.users[username]
-    assert user.roles == []
+    assert [role.name for role in user.roles] == ["user"]
