@@ -4,21 +4,20 @@ import os
 import sys
 from binascii import hexlify
 from unittest import mock
-from urllib.parse import parse_qs
-from urllib.parse import urlparse
+from urllib.parse import parse_qs, urlparse
 
 import pytest
+from bs4 import BeautifulSoup
 from pytest import raises
 from tornado.httputil import url_concat
+from tornado.log import app_log
 
-from .. import orm
-from .. import roles
+from .. import orm, roles, scopes
+from ..roles import roles_to_scopes
 from ..services.auth import _ExpiringDict
 from ..utils import url_path_join
 from .mocking import public_url
-from .test_api import add_user
-from .utils import async_requests
-from .utils import AsyncSession
+from .utils import AsyncSession, async_requests
 
 # mock for sending monotonic counter way into the future
 monotonic_future = mock.patch('time.monotonic', lambda: sys.maxsize)
@@ -204,28 +203,50 @@ async def test_hubauth_service_token(request, app, mockservice_url, scopes, allo
 
 
 @pytest.mark.parametrize(
-    "client_allowed_roles, request_roles, expected_roles",
+    "client_allowed_roles, request_scopes, expected_scopes",
     [
-        # allow empty roles
+        # allow empty permissions
         ([], [], []),
         # allow original 'identify' scope to map to no role
         ([], ["identify"], []),
         # requesting roles outside client list doesn't work
         ([], ["admin"], None),
-        ([], ["token"], None),
-        # requesting nonexistent roles fails in the same way (no server error)
-        ([], ["nosuchrole"], None),
-        # requesting exactly client allow list works
+        ([], ["read:users"], None),
+        # requesting nonexistent roles or scopes fails in the same way (no server error)
+        ([], ["nosuchscope"], None),
+        ([], ["admin:invalid!no=bo!"], None),
+        # requesting role exactly client allow list works
         (["user"], ["user"], ["user"]),
+        # Request individual scope, held by user, not listed in allowed role
         # no explicit request, defaults to all
         (["token", "user"], [], ["token", "user"]),
-        # explicit 'identify' maps to none
-        (["token", "user"], ["identify"], []),
+        # explicit 'identify' maps to read:users:name!user
+        (["token", "user"], ["identify"], ["read:users:name!user=$user"]),
         # any item outside the list isn't allowed
         (["token", "user"], ["token", "server"], None),
+        (["read-only"], ["access:services"], None),
         # requesting subset
         (["admin", "user"], ["user"], ["user"]),
         (["user", "token", "server"], ["token", "user"], ["token", "user"]),
+        (["admin", "user", "read-only"], ["read-only"], ["read-only"]),
+        # Request individual scopes, listed in allowed role
+        (["read-only"], ["access:servers"], ["access:servers"]),
+        # requesting valid subset, some not held by user
+        (
+            ["admin", "user"],
+            ["admin:users", "access:servers", "self"],
+            ["access:servers", "user"],
+        ),
+        (["other"], ["other"], []),
+        # custom scopes
+        (["user"], ["custom:jupyter_server:read:*"], None),
+        (
+            ["read-only"],
+            ["custom:jupyter_server:read:*"],
+            ["custom:jupyter_server:read:*"],
+        ),
+        # this one _should_ work, but doesn't until we implement expanded_scope filtering
+        # (["read-only"], ["custom:jupyter_server:read:*!user=$user"], ["custom:jupyter_server:read:*!user=$user"]),
     ],
 )
 async def test_oauth_service_roles(
@@ -233,8 +254,9 @@ async def test_oauth_service_roles(
     mockservice_url,
     create_user_with_scopes,
     client_allowed_roles,
-    request_roles,
-    expected_roles,
+    request_scopes,
+    expected_scopes,
+    preserve_scopes,
 ):
     service = mockservice_url
     oauth_client = (
@@ -242,31 +264,54 @@ async def test_oauth_service_roles(
         .filter_by(identifier=service.oauth_client_id)
         .one()
     )
-    oauth_client.allowed_roles = [
-        orm.Role.find(app.db, role_name) for role_name in client_allowed_roles
-    ]
+    scopes.define_custom_scopes(
+        {
+            "custom:jupyter_server:read:*": {
+                "description": "read-only access to jupyter server",
+            },
+        },
+    )
+    roles.create_role(
+        app.db,
+        {
+            "name": "read-only",
+            "description": "read-only access to servers",
+            "scopes": [
+                "access:servers",
+                "custom:jupyter_server:read:*",
+            ],
+        },
+    )
+
+    roles.create_role(
+        app.db,
+        {
+            "name": "other",
+            "description": "A role not held by our test user",
+            "scopes": [
+                "admin:users",
+            ],
+        },
+    )
+    oauth_client.allowed_scopes = sorted(
+        roles_to_scopes(
+            [orm.Role.find(app.db, role_name) for role_name in client_allowed_roles]
+        )
+    )
     app.db.commit()
     url = url_path_join(public_url(app, mockservice_url) + 'owhoami/?arg=x')
+    if request_scopes:
+        url = url_concat(url, {"request-scope": " ".join(request_scopes)})
     # first request is only going to login and get us to the oauth form page
     s = AsyncSession()
     user = create_user_with_scopes("access:services")
     roles.grant_role(app.db, user, "user")
+    roles.grant_role(app.db, user, "read-only")
     name = user.name
     s.cookies = await app.login_user(name)
 
     r = await s.get(url)
-    r.raise_for_status()
-    # we should be looking at the oauth confirmation page
-    assert urlparse(r.url).path == app.base_url + 'hub/api/oauth2/authorize'
-    # verify oauth state cookie was set at some point
-    assert set(r.history[0].cookies.keys()) == {'service-%s-oauth-state' % service.name}
-
-    # submit the oauth form to complete authorization
-    data = {}
-    if request_roles:
-        data["scopes"] = request_roles
-    r = await s.post(r.url, data=data, headers={'Referer': r.url})
-    if expected_roles is None:
+    if expected_scopes is None:
         # expected failed auth, stop here
         # verify expected 'invalid scope' error, not server error
         dest_url, _, query = r.url.partition("?")
@@ -274,6 +319,22 @@ async def test_oauth_service_roles(
         assert parse_qs(query).get("error") == ["invalid_scope"]
         assert r.status_code == 400
         return
+
+    r.raise_for_status()
+    # we should be looking at the oauth confirmation page
+    assert urlparse(r.url).path == app.base_url + 'hub/api/oauth2/authorize'
+    # verify oauth state cookie was set at some point
+    assert set(r.history[0].cookies.keys()) == {'service-%s-oauth-state' % service.name}
+
+    page = BeautifulSoup(r.text, "html.parser")
+    scope_inputs = page.find_all("input", {"name": "scopes"})
+    scope_values = [input["value"] for input in scope_inputs]
+    app_log.info(f"Submitting request with scope values {scope_values}")
+    # submit the oauth form to complete authorization
+    data = {}
+    if scope_values:
+        data["scopes"] = scope_values
+    r = await s.post(r.url, data=data, headers={'Referer': r.url})
     r.raise_for_status()
     assert r.url == url
     # verify oauth cookie is set
@@ -285,10 +346,35 @@ async def test_oauth_service_roles(
     r = await s.get(url, allow_redirects=False)
     r.raise_for_status()
     assert r.status_code == 200
+    assert len(r.history) == 0
     reply = r.json()
     sub_reply = {key: reply.get(key, 'missing') for key in ('kind', 'name')}
     assert sub_reply == {'name': user.name, 'kind': 'user'}
 
+    expected_scopes = {s.replace("$user", user.name) for s in expected_scopes}
+
+    # expand roles to scopes (shortcut)
+    for scope in list(expected_scopes):
+        role = orm.Role.find(app.db, scope)
+        if role:
+            expected_scopes.discard(role.name)
+            expected_scopes.update(
+                roles.roles_to_expanded_scopes([role], owner=user.orm_user)
+            )
+
+    if 'inherit' in expected_scopes:
+        expected_scopes = set(scopes.get_scopes_for(user.orm_user))
+
+    # always expect identify/access scopes
+    # on successful authentication
+    expected_scopes.update(scopes.identify_scopes(user.orm_user))
+    expected_scopes.update(scopes.access_scopes(oauth_client))
+    expected_scopes = scopes.reduce_scopes(expected_scopes)
+    have_scopes = scopes.reduce_scopes(set(reply['scopes']))
+    # pytest is better at reporting list differences
+    # than set differences, especially with `-vv`
+    assert sorted(have_scopes) == sorted(expected_scopes)
+
     # token-authenticated request to HubOAuth
     token = app.users[name].new_api_token()
     # token in ?token parameter
@@ -374,7 +460,11 @@ async def test_oauth_access_scopes(
 
 @pytest.mark.parametrize(
     "token_roles, hits_page",
-    [([], True), (['writer'], True), (['writer', 'reader'], False)],
+    [
+        ([], True),
+        (['writer'], True),
+        (['writer', 'reader'], False),
+    ],
 )
 async def test_oauth_page_hit(
     app,
@@ -385,23 +475,30 @@ async def test_oauth_page_hit(
     hits_page,
 ):
     test_roles = {
-        'reader': create_temp_role(['read:users'], role_name='reader'),
-        'writer': create_temp_role(['users:activity'], role_name='writer'),
+        'reader': create_temp_role(['read:users!user'], role_name='reader'),
+        'writer': create_temp_role(['users:activity!user'], role_name='writer'),
     }
     service = mockservice_url
     user = create_user_with_scopes("access:services", "self")
-    user.new_api_token()
-    token = user.api_tokens[0]
-    token.roles = [test_roles[t] for t in token_roles]
+    for role in test_roles.values():
+        roles.grant_role(app.db, user, role)
 
+    # Create a token with the prior authorization
     oauth_client = (
         app.db.query(orm.OAuthClient)
         .filter_by(identifier=service.oauth_client_id)
         .one()
     )
-    oauth_client.allowed_roles = list(test_roles.values())
+    oauth_client.allowed_scopes = sorted(roles_to_scopes(list(test_roles.values())))
+
+    authorized_scopes = roles.roles_to_scopes([test_roles[t] for t in token_roles])
+    authorized_scopes.update(scopes.identify_scopes())
+    authorized_scopes.update(scopes.access_scopes(oauth_client))
+    user.new_api_token(scopes=authorized_scopes)
+    token = user.api_tokens[0]
     token.client_id = service.oauth_client_id
     app.db.commit()
+
     s = AsyncSession()
     s.cookies = await app.login_user(user.name)
     url = url_path_join(public_url(app, service) + 'owhoami/?arg=x')
